from typing import Dict, Any, List, Union
from csle_common.dao.emulation_config.snort_managers_info import SnortIdsManagersInfo
from csle_common.dao.emulation_config.ossec_managers_info import OSSECIDSManagersInfo
from csle_common.dao.emulation_config.kafka_managers_info import KafkaManagersInfo
from csle_common.dao.emulation_config.host_managers_info import HostManagersInfo
from csle_common.dao.emulation_config.five_g_core_managers_info import FiveGCoreManagersInfo
from csle_common.dao.emulation_config.five_g_cu_managers_info import FiveGCUManagersInfo
from csle_common.dao.emulation_config.five_g_du_managers_info import FiveGDUManagersInfo
from csle_common.dao.emulation_config.client_managers_info import ClientManagersInfo
from csle_common.dao.emulation_config.docker_stats_managers_info import DockerStatsManagersInfo
from csle_common.dao.emulation_config.elk_managers_info import ELKManagersInfo
from csle_common.dao.emulation_config.traffic_managers_info import TrafficManagersInfo
from csle_common.dao.emulation_config.ryu_managers_info import RyuManagersInfo
from csle_common.dao.emulation_config.node_container_config import NodeContainerConfig
from csle_common.dao.emulation_config.container_network import ContainerNetwork
from csle_base.json_serializable import JSONSerializable


class EmulationExecutionInfo(JSONSerializable):
    """
    DTO containing the runtime status of an emulation execution
    """

    def __init__(self, emulation_name: str, execution_id: int, snort_ids_managers_info: SnortIdsManagersInfo,
                 ossec_ids_managers_info: OSSECIDSManagersInfo, kafka_managers_info: KafkaManagersInfo,
                 host_managers_info: HostManagersInfo, client_managers_info: ClientManagersInfo,
                 docker_stats_managers_info: DockerStatsManagersInfo, running_containers: List[NodeContainerConfig],
                 stopped_containers: List[NodeContainerConfig], traffic_managers_info: TrafficManagersInfo,
                 active_networks: List[ContainerNetwork],
                 inactive_networks: List[ContainerNetwork], elk_managers_info: ELKManagersInfo,
                 ryu_managers_info: Union[None, RyuManagersInfo], five_g_core_managers_info: FiveGCoreManagersInfo,
                 five_g_cu_managers_info: FiveGCUManagersInfo, five_g_du_managers_info: FiveGDUManagersInfo):
        """
        Initializes the DTO

        :param emulation_name: the name of the emulation
        :param execution_id: the execution ID
        :param snort_ids_managers_info: information about the snort managers
        :param ossec_ids_managers_info: information about the OSSEC managers
        :param kafka_managers_info: information about the Kafka managers
        :param host_managers_info: information about the Host managers
        :param traffic_managers_info: information about the traffic managers
        :param client_managers_info: information about the client managers
        :param docker_stats_managers_info: information about the docker stats managers
        :param running_containers: information about the running containers
        :param stopped_containers: information about the stopped containers
        :param active_networks: information about the active networks
        :param inactive_networks: information about the inactive networks
        :param elk_managers_info: information about the ELK managers
        :param ryu_managers_info: information about the Ryu managers
        :param five_g_core_managers_info: information about the 5G core managers
        :param five_g_cu_managers_info: information about the 5G CU managers
        :param five_g_du_managers_info: information about the 5G DU managers
        """
        self.emulation_name = emulation_name
        self.execution_id = execution_id
        self.snort_ids_managers_info = snort_ids_managers_info
        self.ossec_ids_managers_info = ossec_ids_managers_info
        self.kafka_managers_info = kafka_managers_info
        self.host_managers_info = host_managers_info
        self.client_managers_info = client_managers_info
        self.docker_stats_managers_info = docker_stats_managers_info
        self.running_containers = running_containers
        self.stopped_containers = stopped_containers
        self.active_networks = active_networks
        self.inactive_networks = inactive_networks
        self.elk_managers_info = elk_managers_info
        self.traffic_managers_info = traffic_managers_info
        self.ryu_managers_info = ryu_managers_info
        self.five_g_core_managers_info = five_g_core_managers_info
        self.five_g_cu_managers_info = five_g_cu_managers_info
        self.five_g_du_managers_info = five_g_du_managers_info

    def __str__(self) -> str:
        """
        :return: a string representation of the DTO
        """
        return f"emulation_name: {self.emulation_name}, execution id: {self.execution_id}, " \
               f"snort_ids_managers_info: {self.snort_ids_managers_info}, " \
               f"ossec_ids_managers_info: {self.ossec_ids_managers_info}," \
               f"kafka_managers_info: {self.kafka_managers_info}, host_managers_info: {self.host_managers_info}," \
               f"client_managers_info: {self.client_managers_info}, " \
               f"docker_stats_managers_info: {self.docker_stats_managers_info}, " \
               f"elk_managers_info: {self.elk_managers_info}, traffic_managers_info: {self.traffic_managers_info}" \
               f"running_containers: {list(map(lambda x: str(x), self.running_containers))}, " \
               f"stopped_containers: {list(map(lambda x: str(x), self.stopped_containers))}, " \
               f"active_networks : {list(map(lambda x: str(x), self.active_networks))}, " \
               f"inactive_networks : {list(map(lambda x: str(x), self.inactive_networks))}," \
               f"ryu_managers_info: {self.ryu_managers_info}, " \
               f"five_g_core_managers_info: {self.five_g_core_managers_info}, " \
               f"five_g_cu_managers_info: {self.five_g_cu_managers_info}, " \
               f"five_g_du_managers_info: {self.five_g_du_managers_info}, "

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dictionary representation

        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["emulation_name"] = self.emulation_name
        d["execution_id"] = self.execution_id
        d["snort_ids_managers_info"] = self.snort_ids_managers_info.to_dict()
        d["ossec_ids_managers_info"] = self.ossec_ids_managers_info.to_dict()
        d["kafka_managers_info"] = self.kafka_managers_info.to_dict()
        d["host_managers_info"] = self.host_managers_info.to_dict()
        d["elk_managers_info"] = self.elk_managers_info.to_dict()
        d["traffic_managers_info"] = self.traffic_managers_info.to_dict()
        d["client_managers_info"] = self.client_managers_info.to_dict()
        d["docker_stats_managers_info"] = self.docker_stats_managers_info.to_dict()
        d["running_containers"] = list(map(lambda x: x.to_dict(), self.running_containers))
        d["stopped_containers"] = list(map(lambda x: x.to_dict(), self.stopped_containers))
        d["active_networks"] = list(map(lambda x: x.to_dict(), self.active_networks))
        d["inactive_networks"] = list(map(lambda x: x.to_dict(), self.inactive_networks))
        if self.ryu_managers_info is not None:
            d["ryu_managers_info"] = self.ryu_managers_info.to_dict()
        else:
            d["ryu_managers_info"] = None
        d["five_g_core_managers_info"] = self.five_g_core_managers_info.to_dict()
        d["five_g_cu_managers_info"] = self.five_g_cu_managers_info.to_dict()
        d["five_g_du_managers_info"] = self.five_g_du_managers_info.to_dict()
        return d

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "EmulationExecutionInfo":
        """
        Convert a dict representation to a DTO representation

        :return: a dto representation of the object
        """

        dto = EmulationExecutionInfo(
            emulation_name=d["emulation_name"], execution_id=d["execution_id"],
            snort_ids_managers_info=SnortIdsManagersInfo.from_dict(d["snort_ids_managers_info"]),
            ossec_ids_managers_info=OSSECIDSManagersInfo.from_dict(d["ossec_ids_managers_info"]),
            kafka_managers_info=KafkaManagersInfo.from_dict(d["kafka_managers_info"]),
            elk_managers_info=ELKManagersInfo.from_dict(d["elk_managers_info"]),
            traffic_managers_info=TrafficManagersInfo.from_dict(d["traffic_managers_info"]),
            host_managers_info=HostManagersInfo.from_dict(d["host_managers_info"]),
            client_managers_info=ClientManagersInfo.from_dict(d["client_managers_info"]),
            docker_stats_managers_info=DockerStatsManagersInfo.from_dict(d["docker_stats_managers_info"]),
            running_containers=list(map(lambda x: NodeContainerConfig.from_dict(x), d["running_containers"])),
            stopped_containers=list(map(lambda x: NodeContainerConfig.from_dict(x), d["stopped_containers"])),
            active_networks=list(map(lambda x: ContainerNetwork.from_dict(x), d["active_networks"])),
            inactive_networks=list(map(lambda x: ContainerNetwork.from_dict(x), d["inactive_networks"])),
            ryu_managers_info=RyuManagersInfo.from_dict(d["ryu_managers_info"]),
            five_g_core_managers_info=FiveGCoreManagersInfo.from_dict(d["five_g_core_managers_info"]),
            five_g_cu_managers_info=FiveGCUManagersInfo.from_dict(d["five_g_cu_managers_info"]),
            five_g_du_managers_info=FiveGDUManagersInfo.from_dict(d["five_g_du_managers_info"]))
        return dto

    @staticmethod
    def from_json_file(json_file_path: str) -> "EmulationExecutionInfo":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return EmulationExecutionInfo.from_dict(json.loads(json_str))



import pytest
import logging
import csle_common.constants.constants as constants
from unittest.mock import patch, MagicMock
from csle_common.dao.emulation_config.emulation_env_config import EmulationEnvConfig
from csle_common.controllers.ossec_ids_controller import OSSECIDSController


class TestOssecIdsSuite:
    """
    Test suite for ossec_ids_controller
    """

    @pytest.fixture(autouse=True)
    def emulation_env_config_setup(self) -> None:
        """
        Set up emulation environment configuration

        :return: None
        """
        # setup container configurations
        emulation_env_config = MagicMock(spec=EmulationEnvConfig)
        container1 = MagicMock()
        container1.physical_host_ip = "192.168.1.10"
        container1.docker_gw_bridge_ip = "172.17.0.1"
        container1.get_ips.return_value = ["172.18.0.1"]
        container1.get_full_name.return_value = "container-1"
        container1.name = "container-1"
        container2 = MagicMock()
        container2.physical_host_ip = "192.168.1.10"
        container2.docker_gw_bridge_ip = "172.17.0.2"
        container2.name = "container-2"
        container1.get_ips.return_value = ["172.18.0.2"]
        container1.get_full_name.return_value = "container-2"
        container3 = MagicMock()
        container3.physical_host_ip = "192.168.1.11"
        container3.docker_gw_bridge_ip = "172.17.0.3"
        container3.name = "container-3"
        containers_config = MagicMock()
        emulation_env_config.containers_config = containers_config
        emulation_env_config.containers_config.containers = [container1, container2, container3]
        ossec_ids_manager_config = MagicMock()
        emulation_env_config.ossec_ids_manager_config = ossec_ids_manager_config
        emulation_env_config.ossec_ids_manager_config.ossec_ids_manager_port = 1515
        self.emulation_env_config = emulation_env_config
        self.logger = MagicMock(spec=logging.Logger)
        mock_ossec_ids_status = MagicMock()
        mock_ossec_ids_status.monitor_running = True
        mock_ossec_ids_status.ossec_ids_running = False
        self.mock_ossec_ids_status = mock_ossec_ids_status

    @patch("csle_common.controllers.ossec_ids_controller.OSSECIDSController.stop_ossec_ids")
    def test_stop_ossec_idses(self, mock_stop_ossec_ids) -> None:
        """
        Test the method for stopping the OSSEC IDSes

        :param mock_stop_ossec_ids: mock stop_ossec_ids
        :return: None
        """
        constants.CONTAINER_IMAGES.OSSEC_IDS_IMAGES = [
            self.emulation_env_config.containers_config.containers[0].name,
            self.emulation_env_config.containers_config.containers[1].name
        ]
        physical_host_ip = self.emulation_env_config.containers_config.containers[0].physical_host_ip
        OSSECIDSController.stop_ossec_idses(emulation_env_config=self.emulation_env_config,
                                            physical_host_ip=physical_host_ip, logger=self.logger)
        assert mock_stop_ossec_ids.call_count == 2

    @patch("csle_common.controllers.ossec_ids_controller.OSSECIDSController.start_ossec_ids")
    def test_start_ossec_idses(self, mock_start_ossec_ids) -> None:
        """
        Test the method for starting the OSSEC IDSes

        :param mock_start_ossec_ids: mock start_ossec_ids
        :return: None
        """
        constants.CONTAINER_IMAGES.OSSEC_IDS_IMAGES = [
            self.emulation_env_config.containers_config.containers[0].name,
            self.emulation_env_config.containers_config.containers[1].name
        ]
        physical_server_ip = self.emulation_env_config.containers_config.containers[0].physical_host_ip
        OSSECIDSController.start_ossec_idses(
            emulation_env_config=self.emulation_env_config, physical_server_ip=physical_server_ip, logger=self.logger)
        for i in [0, 1]:
            self.logger.info.assert_any_call(
                f"Starting the OSSEC IDS on ip: "
                f"{self.emulation_env_config.containers_config.containers[i].docker_gw_bridge_ip} "
                f"({self.emulation_env_config.containers_config.containers[i].get_ips()[0]}, "
                f"{self.emulation_env_config.containers_config.containers[i].get_full_name()})")
        assert mock_start_ossec_ids.call_count == 2

    @patch("csle_common.controllers.ossec_ids_controller.OSSECIDSController.start_ossec_ids_manager")
    @patch(
        "csle_common.controllers.ossec_ids_controller.OSSECIDSController"
        ".get_ossec_ids_monitor_thread_status_by_ip_and_port"
    )
    @patch("grpc.insecure_channel")
    def test_start_ossec_ids(
            self, mock_insecure_channel, mock_get_monitor_status, mock_start_manager) -> None:
        """
        Test the method for starting a OSSEC IDS with a specific IP

        :param mock_logger: mock logger
        :param mock_insecure_channel: mock insecure_channel
        :param mock_get_monitor_status: mock get_monitor_status
        :param mock_start_manager: mock start_manager
        :return: None
        """
        mock_get_monitor_status.return_value = self.mock_ossec_ids_status
        mock_channel = MagicMock()
        mock_insecure_channel.return_value.__enter__.return_value = mock_channel
        ip = "192.168.1.10"
        OSSECIDSController.start_ossec_ids(emulation_env_config=self.emulation_env_config, ip=ip, logger=self.logger)
        mock_start_manager.assert_called_once_with(emulation_env_config=self.emulation_env_config, ip=ip,
                                                   logger=self.logger)
        mock_get_monitor_status.assert_called_once_with(
            port=self.emulation_env_config.ossec_ids_manager_config.ossec_ids_manager_port, ip=ip)
        mock_insecure_channel.assert_called_once_with(f"{ip}:1515", options=constants.GRPC_SERVERS.GRPC_OPTIONS)
        self.logger.info.assert_called_once_with(f"Starting OSSEC IDS on {ip}.")

    @patch("csle_common.controllers.ossec_ids_controller.OSSECIDSController.start_ossec_ids_manager")
    @patch(
        "csle_common.controllers.ossec_ids_controller.OSSECIDSController"
        ".get_ossec_ids_monitor_thread_status_by_ip_and_port"
    )
    @patch("grpc.insecure_channel")
    def test_stop_ossec_ids(self, mock_insecure_channel, mock_get_monitor_status, mock_start_manager) -> None:
        """
        Test the method for stopping a OSSEC IDS with a specific IP

        :param mock_insecure_channel: mock insecure_channel
        :param mock_get_monitor_status: mock get_monitor_status
        :param mock_start_manager: mock start_manager
        :return: None
        """
        mock_get_monitor_status.return_value.ossec_ids_running = True
        mock_channel = MagicMock()
        mock_insecure_channel.return_value.__enter__.return_value = mock_channel
        ip = "192.168.1.10"
        OSSECIDSController.start_ossec_ids(emulation_env_config=self.emulation_env_config, ip=ip, logger=self.logger)
        mock_start_manager.assert_called_once_with(emulation_env_config=self.emulation_env_config, ip=ip,
                                                   logger=self.logger)
        mock_get_monitor_status.assert_called_once_with(port=1515, ip=ip)

    @patch("csle_common.controllers.ossec_ids_controller.OSSECIDSController.start_ossec_ids_manager")
    def test_start_ossec_idses_managers(self, mock_start_manager) -> None:
        """
        Test the method for starting OSSEC IDS managers

        :param mock_start_manager: mock start_manager
        :return: None
        """
        constants.CONTAINER_IMAGES = MagicMock()  # type: ignore
        constants.CONTAINER_IMAGES.OSSEC_IDS_IMAGES = [
            self.emulation_env_config.containers_config.containers[0].name,
            self.emulation_env_config.containers_config.containers[1].name
        ]
        physical_server_ip = self.emulation_env_config.containers_config.containers[0].physical_host_ip
        OSSECIDSController.start_ossec_idses_managers(
            emulation_env_config=self.emulation_env_config,
            physical_server_ip=physical_server_ip, logger=self.logger)
        assert mock_start_manager.call_count == 2

    @patch("csle_common.util.emulation_util.EmulationUtil.connect_admin")
    @patch("csle_common.util.emulation_util.EmulationUtil.execute_ssh_cmd")
    @patch("time.sleep", return_value=None)
    def test_start_ossec_ids_manager(self, mock_sleep, mock_execute_ssh_cmd, mock_connect_admin) -> None:
        """
        Test method for starting the OSSEC IDS manager on a specific container

        :param mock_execute_ssh_cmd: mock execute_ssh_cmd
        :param mock_connect_admin: mock connect_admin
        :param mock_sleep: mock sleep
        :return: None
        """
        ip = "192.168.1.10"
        mock_execute_ssh_cmd.return_value = ("ossec_ids_manager running", "", 0)
        constants.COMMANDS.SEARCH_OSSEC_IDS_MANAGER = "ossec_ids_manager running"
        OSSECIDSController.start_ossec_ids_manager(emulation_env_config=self.emulation_env_config, ip=ip,
                                                   logger=self.logger)
        mock_execute_ssh_cmd.assert_called()
        mock_connect_admin.assert_called()

    @patch("csle_common.controllers.ossec_ids_controller.OSSECIDSController.stop_ossec_ids_manager")
    @patch("time.sleep", return_value=None)
    def test_stop_ossec_idses_managers(self, mock_sleep, mock_stop_manager) -> None:
        """
        Test method for stopping ossec ids managers

        :param mock_stop_manager: mock stop_manager
        :param mock_sleep: mock sleep
        :return: None
        """
        constants.CONTAINER_IMAGES.OSSEC_IDS_IMAGES = [
            self.emulation_env_config.containers_config.containers[0].name,
            self.emulation_env_config.containers_config.containers[1].name
        ]
        physical_server_ip = self.emulation_env_config.containers_config.containers[0].physical_host_ip
        OSSECIDSController.stop_ossec_idses_managers(emulation_env_config=self.emulation_env_config,
                                                     physical_server_ip=physical_server_ip, logger=self.logger)
        assert mock_stop_manager.call_count == 2

    @patch("csle_common.util.emulation_util.EmulationUtil.connect_admin")
    @patch("csle_common.util.emulation_util.EmulationUtil.execute_ssh_cmd")
    @patch("time.sleep", return_value=None)
    def test_stop_ossec_ids_manager(self, mock_sleep, mock_execute_ssh_cmd, mock_connect_admin) -> None:
        """
        Test method for stopping the OSSEC IDS manager on a specific container

        :param mock_execute_ssh_cmd: mock execute_ssh_cmd
        :param mock_connect_admin: mock connect_admin
        :param mock_sleep: mock sleep
        :return: None
        """
        ip = "192.168.1.10"
        mock_execute_ssh_cmd.return_value = ("ossec_ids_manager running", "", 0)
        OSSECIDSController.stop_ossec_ids_manager(emulation_env_config=self.emulation_env_config, ip=ip,
                                                  logger=self.logger)
        mock_execute_ssh_cmd.assert_called()
        mock_connect_admin.assert_called()

    @patch("csle_common.controllers.ossec_ids_controller.OSSECIDSController.start_ossec_idses_managers")
    @patch("csle_common.controllers.ossec_ids_controller.OSSECIDSController.start_ossec_ids_monitor_thread")
    def test_start_ossec_idses_monitor_threads(self, mock_start_monitor, mock_start_managers) -> None:
        """
        Unit test for the start_ossec_ids_monnitor_threads method of the OSSEC IDS controller

        :param mock_start_monitor: mock of the start_monitor method
        :param mock_start_managers: mock of the start_managers method
        :return: None
        """
        constants.CONTAINER_IMAGES.OSSEC_IDS_IMAGES = ["container-1", "container-2"]
        physical_server_ip = "192.168.1.10"
        OSSECIDSController.start_ossec_idses_monitor_threads(emulation_env_config=self.emulation_env_config,
                                                             physical_server_ip=physical_server_ip,
                                                             logger=self.logger)
        mock_start_managers.assert_called_once_with(emulation_env_config=self.emulation_env_config,
                                                    physical_server_ip=physical_server_ip, logger=self.logger)
        assert mock_start_monitor.call_count == 2

    @patch("csle_common.controllers.ossec_ids_controller.OSSECIDSController.start_ossec_ids_manager")
    @patch(
        "csle_common.controllers.ossec_ids_controller.OSSECIDSController"
        ".get_ossec_ids_monitor_thread_status_by_ip_and_port"
    )
    @patch("grpc.insecure_channel")
    def test_start_ossec_ids_monitor_thread(self, mock_insecure_channel, mock_get_monitor_status, mock_start_manager) \
            -> None:
        """
        A method that sends a request to the OSSECIDSManager on a specific IP to start
        the IDS manager and the monitor thread

        :param mock_insecure_channel: mock insecure_channel
        :param mock_get_monitor_status: mock get_monitor_status
        :param mock_start_manager:  mock start_manager
        :return: None
        """
        mock_get_monitor_status.return_value.ossec_ids_running = False
        mock_channel = MagicMock()
        mock_insecure_channel.return_value.__enter__.return_value = mock_channel
        ip = "192.168.1.10"
        OSSECIDSController.start_ossec_ids_monitor_thread(emulation_env_config=self.emulation_env_config, ip=ip,
                                                          logger=self.logger)
        mock_start_manager.assert_called_once_with(emulation_env_config=self.emulation_env_config, ip=ip,
                                                   logger=self.logger)
        mock_get_monitor_status.assert_called_once_with(port=1515, ip=ip)

    @patch("csle_common.controllers.ossec_ids_controller.OSSECIDSController.stop_ossec_ids_monitor_thread")
    def test_stop_ossec_ids_monitor_threads(self, mock_stop_monitor) -> None:
        """
        Test a method that sends a request to the OSSECIDSManager on every container that runs
        an IDS to stop the monitor threads

        :param mock_stop_monitor: _description_
        :return: None
        """
        constants.CONTAINER_IMAGES = MagicMock()  # type: ignore
        constants.CONTAINER_IMAGES.OSSEC_IDS_IMAGES = ["container-1", "container-2"]
        physical_server_ip = "192.168.1.10"
        OSSECIDSController.stop_ossec_idses_monitor_threads(
            emulation_env_config=self.emulation_env_config, physical_server_ip=physical_server_ip, logger=self.logger)
        assert mock_stop_monitor.call_count == 2

    @patch("csle_common.controllers.ossec_ids_controller.OSSECIDSController.start_ossec_ids_manager")
    @patch("grpc.insecure_channel")
    def test_stop_ossec_ids_monitor_thread(self, mock_insecure_channel, mock_start_manager) -> None:
        """
        Test a method that sends a request to the OSSECIDSManager for a specific IP to stop the monitor thread

        :param mock_start_manager: mock start_manager
        :param mock_insecure_channel: mock of the channel
        :return: None
        """
        mock_channel = MagicMock()
        mock_insecure_channel.return_value.__enter__.return_value = mock_channel
        ip = "192.168.1.10"
        OSSECIDSController.stop_ossec_ids_monitor_thread(emulation_env_config=self.emulation_env_config, ip=ip,
                                                         logger=self.logger)
        mock_start_manager.assert_called_once_with(emulation_env_config=self.emulation_env_config, ip=ip,
                                                   logger=self.logger)

    @patch("csle_common.controllers.ossec_ids_controller.OSSECIDSController.start_ossec_idses_managers")
    @patch("grpc.insecure_channel")
    def test_get_ossec_idses_monitor_threads_statuses(self, mock_insecure_channel, mock_start_managers) -> None:
        """
        Test a method that sends a request to the OSSECIDSManager on every container to get the status of the
        IDS monitor thread

        :param mock_start_manager: mock start_manager
        :param mock_insecure_channel: mock of the channel
        :return: None
        """
        mock_channel = MagicMock()
        mock_insecure_channel.return_value.__enter__.return_value = mock_channel
        physical_server_ip = "192.168.1.10"
        constants.CONTAINER_IMAGES = MagicMock()  # type: ignore
        constants.CONTAINER_IMAGES.OSSEC_IDS_IMAGES = ["container-1", "container-2"]
        OSSECIDSController.get_ossec_idses_monitor_threads_statuses(
            emulation_env_config=self.emulation_env_config, physical_server_ip=physical_server_ip, logger=self.logger)
        mock_start_managers.assert_called_once_with(
            emulation_env_config=self.emulation_env_config, physical_server_ip=physical_server_ip, logger=self.logger)

    def test_get_ossec_idses_managers_ips(self) -> None:
        """
        Tests a method that extracts the IPS of the OSSEC IDS managers in a given emulation

        :return: None
        """
        constants.CONTAINER_IMAGES = MagicMock()  # type: ignore
        constants.CONTAINER_IMAGES.OSSEC_IDS_IMAGES = ["container-1", "container-2"]
        ips = OSSECIDSController.get_ossec_idses_managers_ips(emulation_env_config=self.emulation_env_config)
        assert ips == ["172.17.0.1", "172.17.0.2"]

    def test_get_ossec_idses_managers_ports(self) -> None:
        """
        Tests a method that extracts the ports of the OSSEC IDS managers in a given emulation

        :return: None
        """
        constants.CONTAINER_IMAGES = MagicMock()  # type: ignore
        constants.CONTAINER_IMAGES.OSSEC_IDS_IMAGES = ["container-1", "container-2"]
        ports = OSSECIDSController.get_ossec_idses_managers_ports(emulation_env_config=self.emulation_env_config)
        assert ports == [1515, 1515]

    @patch("grpc.insecure_channel")
    def test_get_ossec_ids_monitor_thread_status_by_ip_and_port(self, mock_insecure_channel) -> None:
        """
        Tests a method that sends a request to the OSSECIDSManager with a specific port and ip
        to get the status of the IDS monitor thread

        :param mock_insecure_channel: mock insecure_channel
        :return: None
        """
        port = 1515
        ip = "192.168.1.100"
        mock_channel = MagicMock()
        mock_insecure_channel.return_value.__enter__.return_value = mock_channel
        OSSECIDSController.get_ossec_ids_monitor_thread_status_by_ip_and_port(port=port, ip=ip)
        mock_insecure_channel.assert_called_once_with(f"{ip}:{port}", options=constants.GRPC_SERVERS.GRPC_OPTIONS)

    @patch("csle_common.controllers.ossec_ids_controller.OSSECIDSController.get_ossec_idses_managers_ips")
    @patch("csle_common.controllers.ossec_ids_controller.OSSECIDSController.get_ossec_idses_managers_ports")
    @patch(
        "csle_common.controllers.ossec_ids_controller.OSSECIDSController"
        ".get_ossec_ids_monitor_thread_status_by_ip_and_port"
    )
    @patch("csle_common.util.emulation_util.EmulationUtil.physical_ip_match")
    @patch("csle_collector.ossec_ids_manager.ossec_ids_manager_util.OSSecManagerUtil.ossec_ids_monitor_dto_empty")
    def test_get_ossec_managers_info(
            self, mock_ossec_dto_empty, mock_physical_ip_match, mock_get_status, mock_get_ports, mock_get_ips) -> None:
        """
        Tests the method that extracts the information of the OSSEC IDS managers for a given emulation

        :param mock_ossec_dto_empty: mock ossec_dto_empty
        :param mock_physical_ip_match: mock physical_ip_match
        :param mock_get_status: mock get_status
        :param mock_get_ports: mock get_ports
        :param mock_get_ips: mock get_ips
        :retun: None
        """
        emulation_env_config = self.emulation_env_config
        emulation_env_config.execution_id = "test_execution_id"
        emulation_env_config.name = "test_emulation_name"
        emulation_env_config.ossec_ids_manager_config.ossec_ids_manager_port = 1515
        active_ips = ["192.168.1.10", "192.168.1.11"]
        physical_host_ip = "192.168.1.1"
        logger = self.logger
        mock_get_ips.return_value = ["192.168.1.10", "192.168.1.12"]
        mock_get_ports.return_value = [1515, 1515]
        mock_status_1 = MagicMock()
        mock_get_status.side_effect = [mock_status_1, Exception("Test Exception")]
        mock_physical_ip_match.side_effect = [True, True]
        mock_ossec_dto_empty.return_value = MagicMock()
        OSSECIDSController.get_ossec_managers_info(emulation_env_config, active_ips, logger, physical_host_ip)
        mock_get_ips.assert_called_once_with(emulation_env_config=emulation_env_config)
        mock_get_ports.assert_called_once_with(emulation_env_config=emulation_env_config)
        assert mock_get_status.call_count == 1



from typing import Dict
from csle_common.dao.simulation_config.action import Action
from csle_common.dao.simulation_config.action_space_config import ActionSpaceConfig
from csle_common.dao.simulation_config.agent_log import AgentLog
from csle_common.dao.simulation_config.env_parameter import EnvParameter
from csle_common.dao.simulation_config.env_parameters_config import EnvParametersConfig
from csle_common.dao.simulation_config.initial_state_distribution_config import InitialStateDistributionConfig
from csle_common.dao.simulation_config.joint_action_space_config import JointActionSpaceConfig
from csle_common.dao.simulation_config.joint_observation_space_config import JointObservationSpaceConfig
from csle_common.dao.simulation_config.observation_space_config import ObservationSpaceConfig
from csle_common.dao.simulation_config.observation import Observation
from csle_common.dao.simulation_config.observation_function_config import ObservationFunctionConfig
from csle_common.dao.simulation_config.player_config import PlayerConfig
from csle_common.dao.simulation_config.players_config import PlayersConfig
from csle_common.dao.simulation_config.reward_function_config import RewardFunctionConfig
from csle_common.dao.simulation_config.simulation_env_config import SimulationEnvConfig
from csle_common.dao.simulation_config.state_space_config import StateSpaceConfig
from csle_common.dao.simulation_config.state import State
from csle_common.dao.simulation_config.transition_operator_config import TransitionOperatorConfig
from csle_common.dao.simulation_config.simulation_trace import SimulationTrace
from csle_common.dao.simulation_config.simulation_env_input_config import SimulationEnvInputConfig


class ExampleInputConfig(SimulationEnvInputConfig):
    """
    Test instance of the abstract SimulationEnvInputConfig class
    """

    def __init__(self, x: int) -> None:
        """
        Initializes the object

        :param x: the input parameter
        :return: None
        """
        self.x = x

    def to_dict(self) -> Dict[str, int]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d: Dict[str, int] = {}
        d["x"] = self.x
        return d

    @staticmethod
    def from_dict(d: Dict[str, int]) -> "ExampleInputConfig":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = ExampleInputConfig(x=d["x"])
        return obj

    @staticmethod
    def from_json_file(json_file_path: str) -> "ExampleInputConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return ExampleInputConfig.from_dict(json.loads(json_str))


class TestSimulationConfigDaoSuite:
    """
    Test suite for simulation config data access objects (DAOs)
    """

    def test_action(self, example_action: Action) -> None:
        """
        Tests creation and dict conversion of the Action DAO

        :param example_action: an example Action
        :return: None
        """
        assert isinstance(example_action.to_dict(), dict)
        assert isinstance(Action.from_dict(example_action.to_dict()), Action)
        assert Action.from_dict(example_action.to_dict()).to_dict() == example_action.to_dict()
        assert Action.from_dict(example_action.to_dict()) == example_action

    def test_action_space_config(self, example_action_space_config: ActionSpaceConfig) -> None:
        """
        Tests creation and dict conversion of the ActionSpaceConfig DAO

        :param example_action_space_config: an example ActionSpaceConfig
        :return: None
        """
        assert isinstance(example_action_space_config.to_dict(), dict)
        assert isinstance(ActionSpaceConfig.from_dict(example_action_space_config.to_dict()),
                          ActionSpaceConfig)
        assert ActionSpaceConfig.from_dict(example_action_space_config.to_dict()).to_dict() == \
               example_action_space_config.to_dict()
        assert ActionSpaceConfig.from_dict(example_action_space_config.to_dict()) == example_action_space_config

    def test_agent_log(self, example_agent_log: AgentLog) -> None:
        """
        Tests creation and dict conversion of the AgentLog DAO

        :param example_agent_log: an example AgentLog
        :return: None
        """
        assert isinstance(example_agent_log.to_dict(), dict)
        assert isinstance(AgentLog.from_dict(example_agent_log.to_dict()), AgentLog)
        assert AgentLog.from_dict(example_agent_log.to_dict()).to_dict() == example_agent_log.to_dict()
        assert AgentLog.from_dict(example_agent_log.to_dict()) == example_agent_log

    def test_observation(self, example_observation: Observation) -> None:
        """
        Tests creation and dict conversion of the Observation DAO

        :param example_observation: an example Observation
        :return: None
        """
        assert isinstance(example_observation.to_dict(), dict)
        assert isinstance(Observation.from_dict(example_observation.to_dict()), Observation)
        assert Observation.from_dict(example_observation.to_dict()).to_dict() == example_observation.to_dict()
        assert Observation.from_dict(example_observation.to_dict()) == example_observation

    def test_env_parameter(self, example_env_parameter: EnvParameter) -> None:
        """
        Tests creation and dict conversion of the EnvParameter DAO

        :param example_env_parameter: an example EnvParameter
        :return: None
        """
        assert isinstance(example_env_parameter.to_dict(), dict)
        assert isinstance(EnvParameter.from_dict(example_env_parameter.to_dict()), EnvParameter)
        assert EnvParameter.from_dict(example_env_parameter.to_dict()).to_dict() == example_env_parameter.to_dict()
        assert EnvParameter.from_dict(example_env_parameter.to_dict()) == example_env_parameter

    def test_env_parameters_config(self, example_env_parameters_config: EnvParametersConfig) -> None:
        """
        Tests creation and dict conversion of the EnvParametersConfig DAO

        :param example_env_parameters_config: an example EnvParametersConfig
        :return: None
        """
        assert isinstance(example_env_parameters_config.to_dict(), dict)
        assert isinstance(EnvParametersConfig.from_dict(example_env_parameters_config.to_dict()), EnvParametersConfig)
        assert (EnvParametersConfig.from_dict(example_env_parameters_config.to_dict()).to_dict() ==
                example_env_parameters_config.to_dict())
        assert EnvParametersConfig.from_dict(example_env_parameters_config.to_dict()) == example_env_parameters_config

    def test_initial_state_distribution_config(
            self, example_initial_state_distribution_config: InitialStateDistributionConfig) -> None:
        """
        Tests creation and dict conversion of the InitialStateDistributionConfig DAO

        :param example_initial_state_distribution_config: an example InitialStateDistributionConfig
        :return: None
        """
        assert isinstance(example_initial_state_distribution_config.to_dict(), dict)
        assert isinstance(InitialStateDistributionConfig.from_dict(example_initial_state_distribution_config.to_dict()),
                          InitialStateDistributionConfig)
        assert (InitialStateDistributionConfig.from_dict(example_initial_state_distribution_config.to_dict()).to_dict()
                == example_initial_state_distribution_config.to_dict())
        assert (InitialStateDistributionConfig.from_dict(example_initial_state_distribution_config.to_dict()) ==
                example_initial_state_distribution_config)

    def test_joint_action_space_config(self, example_joint_action_space_config: JointActionSpaceConfig) -> None:
        """
        Tests creation and dict conversion of the JointActionSpaceConfig DAO

        :param example_joint_action_space_config: an example JointActionSpaceConfig
        :return: None
        """
        assert isinstance(example_joint_action_space_config.to_dict(), dict)
        assert isinstance(JointActionSpaceConfig.from_dict(example_joint_action_space_config.to_dict()),
                          JointActionSpaceConfig)
        assert (JointActionSpaceConfig.from_dict(example_joint_action_space_config.to_dict()).to_dict() ==
                example_joint_action_space_config.to_dict())
        assert (JointActionSpaceConfig.from_dict(example_joint_action_space_config.to_dict()) ==
                example_joint_action_space_config)

    def test_observation_space_config(self, example_observation_space_config: ObservationSpaceConfig) -> None:
        """
        Tests creation and dict conversion of the ObservationSpaceConfig DAO

        :param example_observation_space_config: an example ObservationSpaceConfig
        :return: None
        """
        assert isinstance(example_observation_space_config.to_dict(), dict)
        assert isinstance(ObservationSpaceConfig.from_dict(example_observation_space_config.to_dict()),
                          ObservationSpaceConfig)
        assert (ObservationSpaceConfig.from_dict(example_observation_space_config.to_dict()).to_dict() ==
                example_observation_space_config.to_dict())
        assert (ObservationSpaceConfig.from_dict(example_observation_space_config.to_dict()) ==
                example_observation_space_config)

    def test_joint_observation_space_config(
            self, example_joint_observation_space_config: JointObservationSpaceConfig) -> None:
        """
        Tests creation and dict conversion of the JointObservationSpaceConfig DAO

        :param example_joint_observation_space_config: an example JointObservationSpaceConfig
        :return: None
        """
        assert isinstance(example_joint_observation_space_config.to_dict(), dict)
        assert isinstance(JointObservationSpaceConfig.from_dict(example_joint_observation_space_config.to_dict()),
                          JointObservationSpaceConfig)
        assert (JointObservationSpaceConfig.from_dict(example_joint_observation_space_config.to_dict()).to_dict() ==
                example_joint_observation_space_config.to_dict())
        assert (JointObservationSpaceConfig.from_dict(example_joint_observation_space_config.to_dict()) ==
                example_joint_observation_space_config)

    def test_observation_function_config(self, example_observation_function_config: ObservationFunctionConfig) -> None:
        """
        Tests creation and dict conversion of the ObservationFunctionConfig DAO

        :param example_observation_function_config: an example ObservationFunctionConfig
        :return: None
        """
        assert isinstance(example_observation_function_config.to_dict(), dict)
        assert isinstance(ObservationFunctionConfig.from_dict(example_observation_function_config.to_dict()),
                          ObservationFunctionConfig)
        assert (ObservationFunctionConfig.from_dict(example_observation_function_config.to_dict()).to_dict() ==
                example_observation_function_config.to_dict())
        assert (ObservationFunctionConfig.from_dict(example_observation_function_config.to_dict()) ==
                example_observation_function_config)

    def test_player_config(self, example_player_config: PlayerConfig) -> None:
        """
        Tests creation and dict conversion of the PlayerConfig DAO

        :param example_player_config: an example PlayerConfig
        :return: None
        """
        assert isinstance(example_player_config.to_dict(), dict)
        assert isinstance(PlayerConfig.from_dict(example_player_config.to_dict()), PlayerConfig)
        assert PlayerConfig.from_dict(example_player_config.to_dict()).to_dict() == example_player_config.to_dict()
        assert PlayerConfig.from_dict(example_player_config.to_dict()) == example_player_config

    def test_players_config(self, example_players_config: PlayersConfig) -> None:
        """
        Tests creation and dict conversion of the PlayersConfig DAO

        :param example_players_config: an example PlayersConfig
        :return: None
        """
        assert isinstance(example_players_config.to_dict(), dict)
        assert isinstance(PlayersConfig.from_dict(example_players_config.to_dict()), PlayersConfig)
        assert PlayersConfig.from_dict(example_players_config.to_dict()).to_dict() == example_players_config.to_dict()
        assert PlayersConfig.from_dict(example_players_config.to_dict()) == example_players_config

    def test_reward_function_config(self, example_reward_function_config: RewardFunctionConfig) -> None:
        """
        Tests creation and dict conversion of the RewardFunctionConfig DAO

        :param example_reward_function_config: an example RewardFunctionConfig
        :return: None
        """
        assert isinstance(example_reward_function_config.to_dict(), dict)
        assert isinstance(RewardFunctionConfig.from_dict(example_reward_function_config.to_dict()),
                          RewardFunctionConfig)
        assert (RewardFunctionConfig.from_dict(example_reward_function_config.to_dict()).to_dict() ==
                example_reward_function_config.to_dict())
        assert RewardFunctionConfig.from_dict(example_reward_function_config.to_dict()) == \
               example_reward_function_config

    def test_simulation_env_input_config(self, example_input_config: ExampleInputConfig) -> None:
        """
        Tests creation and dict conversion of the SimulationEnvInputConfig DAO

        :param example_input_config: an example ExampleInputConfig
        :return: None
        """
        assert isinstance(example_input_config.to_dict(), dict)
        assert isinstance(ExampleInputConfig.from_dict(example_input_config.to_dict()),
                          ExampleInputConfig)
        assert (ExampleInputConfig.from_dict(example_input_config.to_dict()).to_dict() ==
                example_input_config.to_dict())
        assert ExampleInputConfig.from_dict(example_input_config.to_dict()) == example_input_config

    def test_simulation_trace(self, example_simulation_trace: SimulationTrace) -> None:
        """
        Tests creation and dict conversion of the SimulationTrace DAO

        :param example_simulation_trace: an example SimulationTrace
        :return: None
        """
        assert isinstance(example_simulation_trace.to_dict(), dict)
        assert isinstance(SimulationTrace.from_dict(example_simulation_trace.to_dict()), SimulationTrace)
        assert SimulationTrace.from_dict(example_simulation_trace.to_dict()).to_dict() == \
               example_simulation_trace.to_dict()
        assert SimulationTrace.from_dict(example_simulation_trace.to_dict()) == example_simulation_trace

    def test_state(self, example_state: State) -> None:
        """
        Tests creation and dict conversion of the State DAO

        :param example_state: an example State
        :return: None
        """
        assert isinstance(example_state.to_dict(), dict)
        assert isinstance(State.from_dict(example_state.to_dict()), State)
        assert State.from_dict(example_state.to_dict()).to_dict() == example_state.to_dict()
        assert State.from_dict(example_state.to_dict()) == example_state

    def test_state_space_config(self, example_state_space_config: StateSpaceConfig) -> None:
        """
        Tests creation and dict conversion of the StateSpaceConfig DAO

        :param example_state_space_config: an example StateSpaceConfig
        :return: None
        """
        assert isinstance(example_state_space_config.to_dict(), dict)
        assert isinstance(StateSpaceConfig.from_dict(example_state_space_config.to_dict()), StateSpaceConfig)
        assert StateSpaceConfig.from_dict(example_state_space_config.to_dict()).to_dict() == \
               example_state_space_config.to_dict()
        assert StateSpaceConfig.from_dict(example_state_space_config.to_dict()) == example_state_space_config

    def test_transition_operator_config(self, example_transition_operator_config: TransitionOperatorConfig) -> None:
        """
        Tests creation and dict conversion of the TransitionOperatorConfig DAO

        :param example_transition_operator_config: an example TransitionOperatorConfig
        :return: None
        """
        assert isinstance(example_transition_operator_config.to_dict(), dict)
        assert isinstance(TransitionOperatorConfig.from_dict(example_transition_operator_config.to_dict()),
                          TransitionOperatorConfig)
        assert (TransitionOperatorConfig.from_dict(example_transition_operator_config.to_dict()).to_dict() ==
                example_transition_operator_config.to_dict())
        assert (TransitionOperatorConfig.from_dict(example_transition_operator_config.to_dict()) ==
                example_transition_operator_config)

    def test_simulation_env_config(self, example_simulation_env_config: SimulationEnvConfig) -> None:
        """
        Tests creation and dict conversion of the SimulationEnvConfig DAO

        :param example_simulation_env_config: an example SimulationEnvConfig
        :return: None
        """
        assert isinstance(example_simulation_env_config.to_dict(), dict)



from csle_common.dao.emulation_observation.common.emulation_port_observation_state import EmulationPortObservationState
from csle_common.dao.emulation_observation.attacker. \
    emulation_attacker_machine_observation_state import EmulationAttackerMachineObservationState
from csle_common.dao.emulation_observation.attacker. \
    emulation_attacker_observation_state import EmulationAttackerObservationState
from csle_common.dao.emulation_observation.common.emulation_connection_observation_state \
    import EmulationConnectionObservationState
from csle_common.dao.emulation_observation.common.emulation_vulnerability_observation_state \
    import EmulationVulnerabilityObservationState
from csle_common.dao.emulation_observation.defender.emulation_defender_machine_observation_state import \
    EmulationDefenderMachineObservationState
from csle_common.dao.emulation_observation.defender.emulation_defender_observation_state import \
    EmulationDefenderObservationState


class TestEmulationObservationDaoSuite:
    """
    Test suite for emulation observation data access objects (DAOs)
    """

    def test_emulation_port_observation_state(
            self, example_emulation_port_observation_state: EmulationPortObservationState) -> None:
        """
        Tests creation and dict conversion of the EmulationPortObservationState DAO

        :param example_emulation_port_observation_state: an example EmulationPortObservationState
        :return: None
        """
        assert isinstance(example_emulation_port_observation_state.to_dict(), dict)
        assert isinstance(EmulationPortObservationState.from_dict(example_emulation_port_observation_state.to_dict()),
                          EmulationPortObservationState)
        assert (EmulationPortObservationState.from_dict(example_emulation_port_observation_state.to_dict()).to_dict() ==
                example_emulation_port_observation_state.to_dict())
        assert (EmulationPortObservationState.from_dict(example_emulation_port_observation_state.to_dict()) ==
                example_emulation_port_observation_state)

    def test_emulation_attacker_machine_observation_state(
            self, example_emulation_attacker_machine_observation_state: EmulationAttackerObservationState) -> None:
        """
        Tests creation and dict conversion of the EmulationAttackerMachineObservationState DAO

        :param example_emulation_attacker_machine: an example EmulationAttackerMachineObservationState
        :return: None
        """
        assert isinstance(example_emulation_attacker_machine_observation_state.to_dict(), dict)
        assert isinstance(EmulationAttackerMachineObservationState.from_dict
                          (example_emulation_attacker_machine_observation_state.to_dict()),
                          EmulationAttackerMachineObservationState)
        assert (EmulationAttackerMachineObservationState.from_dict
                (example_emulation_attacker_machine_observation_state.to_dict()).to_dict() ==
                example_emulation_attacker_machine_observation_state.to_dict())
        assert (EmulationAttackerMachineObservationState.from_dict
                (example_emulation_attacker_machine_observation_state.to_dict()) ==
                example_emulation_attacker_machine_observation_state)

    def test_emulation_attacker_observation_state(
            self, example_emulation_attacker_observation_state: EmulationAttackerObservationState) -> None:
        """
        Tests creation and dict conversion of the EmulationAttackerObservationState DAO

        :param example_emulation_attacker_observation_state: an example EmulationAttackerObservationState
        :return: None
        """
        assert isinstance(example_emulation_attacker_observation_state.to_dict(), dict)
        assert isinstance(EmulationAttackerObservationState.from_dict
                          (example_emulation_attacker_observation_state.to_dict()), EmulationAttackerObservationState)
        assert (EmulationAttackerObservationState.from_dict
                (example_emulation_attacker_observation_state.to_dict()).to_dict() ==
                example_emulation_attacker_observation_state.to_dict())
        assert (EmulationAttackerObservationState.from_dict
                (example_emulation_attacker_observation_state.to_dict()) ==
                example_emulation_attacker_observation_state)

    def test_emulation_connection_observation_state(
            self, example_emulation_connection_observation_state: EmulationConnectionObservationState) -> None:
        """
        Tests creation and dict conversion of the EmulationConnectionObservationState DAO

        :param example_emulation_connection_observation_state: an example EmulationConnectionObservationState
        :return: None
        """
        assert isinstance(example_emulation_connection_observation_state.to_dict(), dict)
        assert isinstance(EmulationConnectionObservationState.from_dict
                          (example_emulation_connection_observation_state.to_dict()),
                          EmulationConnectionObservationState)
        assert (EmulationConnectionObservationState.from_dict
                (example_emulation_connection_observation_state.to_dict()).to_dict() ==
                example_emulation_connection_observation_state.to_dict())
        assert (EmulationConnectionObservationState.from_dict
                (example_emulation_connection_observation_state.to_dict()) ==
                example_emulation_connection_observation_state)

    def test_emulation_vulnerability_observation_state(
            self, example_emulation_vulnerability_observation_state: EmulationVulnerabilityObservationState) -> None:
        """
        Tests creation and dict conversion of the EmulationVulnerabilityObservationState DAO

        :param example_emulation_vulnerability_observation_state: an example EmulationVulnerabilityObservationState
        :return: None
        """
        assert isinstance(example_emulation_vulnerability_observation_state.to_dict(), dict)
        assert isinstance(EmulationVulnerabilityObservationState.from_dict
                          (example_emulation_vulnerability_observation_state.to_dict()),
                          EmulationVulnerabilityObservationState)
        assert (EmulationVulnerabilityObservationState.from_dict
                (example_emulation_vulnerability_observation_state.to_dict()).to_dict() ==
                example_emulation_vulnerability_observation_state.to_dict())
        assert (EmulationVulnerabilityObservationState.from_dict
                (example_emulation_vulnerability_observation_state.to_dict()) ==
                example_emulation_vulnerability_observation_state)

    def test_emulation_defender_machine_observation_state(
            self,
            example_emulation_defender_machine_observation_state: EmulationDefenderMachineObservationState) -> None:
        """
        Tests creation and dict conversion of the EmulationDefenderMachineObservationState DAO

        :param example_emulation_defender_machine_observation_state: an example EmulationDefenderMachineObservationState
        :return: None
        """
        assert isinstance(example_emulation_defender_machine_observation_state.to_dict(), dict)
        assert isinstance(EmulationDefenderMachineObservationState.from_dict
                          (example_emulation_defender_machine_observation_state.to_dict()),
                          EmulationDefenderMachineObservationState)
        assert (EmulationDefenderMachineObservationState.from_dict
                (example_emulation_defender_machine_observation_state.to_dict()).to_dict() ==
                example_emulation_defender_machine_observation_state.to_dict())
        assert (EmulationDefenderMachineObservationState.from_dict
                (example_emulation_defender_machine_observation_state.to_dict()) ==
                example_emulation_defender_machine_observation_state)

    def test_emulation_defender_observation_state(
            self, example_emulation_defender_observation_state: EmulationDefenderObservationState) -> None:
        """
        Tests creation and dict conversion of the EmulationDefenderObservationState DAO

        :param example_emulation_defender_observation_state: an example EmulationDefenderObservationState
        :return: None
        """
        assert isinstance(example_emulation_defender_observation_state.to_dict(), dict)
        assert isinstance(EmulationDefenderObservationState.from_dict(
            example_emulation_defender_observation_state.to_dict()),
            EmulationDefenderObservationState)
        assert (EmulationDefenderObservationState.from_dict
                (example_emulation_defender_observation_state.to_dict()).to_dict() ==
                example_emulation_defender_observation_state.to_dict())
        assert (EmulationDefenderObservationState.from_dict(example_emulation_defender_observation_state.to_dict()) ==
                example_emulation_defender_observation_state)



import torch
from csle_common.models.fnn_w_softmax import FNNwithSoftmax
from csle_common.models.ppo_network import PPONetwork
from csle_common.models.q_network import QNetwork
from csle_common.dao.training.agent_type import AgentType


class TestModelsSuite:
    """
    Test suite for the models package
    """

    def test_fnn_w_softmax(self) -> None:
        """
        Tests creation of the FNN-with-softmax model

        :return: None
        """
        output_dim = 1
        input_dim = 10
        hidden_dim = 64
        num_hidden_layers = 4
        hidden_activation = "ReLU"
        model = FNNwithSoftmax(input_dim=input_dim, output_dim=output_dim, hidden_dim=hidden_dim,
                               num_hidden_layers=num_hidden_layers, hidden_activation=hidden_activation)
        assert model is not None
        assert model.output_dim == output_dim
        assert model.input_dim == input_dim
        assert model.num_hidden_layers == num_hidden_layers
        assert model.hidden_activation == hidden_activation
        assert len(model.layers) == (num_hidden_layers + 2) * 2

        y = model.forward(torch.tensor([[1.0] * input_dim]))
        assert y is not None
        assert len(y.detach().numpy()[0]) == output_dim

    def test_ppo_networK(self) -> None:
        """
        Tests creation of the PPO network model

        :return: None
        """
        output_dim_critic = 1
        output_dim_action = 1
        input_dim = 10
        hidden_dim = 64
        num_hidden_layers = 4
        std_critic = 1.0
        std_action = 0.01
        model = PPONetwork(input_dim=input_dim, output_dim_critic=output_dim_critic,
                           num_hidden_layers=num_hidden_layers, hidden_layer_dim=hidden_dim, std_critic=std_critic,
                           std_action=std_action, output_dim_action=output_dim_action)
        assert model is not None
        assert model.input_dim == input_dim
        assert model.output_dim_critic == output_dim_critic
        assert model.output_dim_action == output_dim_action
        assert model.std_critic == std_critic
        assert model.std_action == std_action
        assert model.num_hidden_layers == num_hidden_layers
        assert model.hidden_layer_dim == hidden_dim
        assert len(list(model.critic.modules())) == (num_hidden_layers + 3)
        assert len(list(model.aux_critic.modules())) == (num_hidden_layers + 3)
        assert len(list(model.actor.modules())) == (num_hidden_layers + 3)
        y = model.get_value(x=torch.tensor([[1.0] * input_dim]))
        assert y is not None
        assert len(y.detach().numpy()[0]) == output_dim_critic
        y2 = model.get_pi(x=torch.tensor([[1.0] * input_dim]))
        assert y2 is not None
        assert len(y2.probs.detach().numpy()[0]) == output_dim_critic

    def test_q_network(self) -> None:
        """
        Tests creation of the Q-network model

        :return: None
        """
        output_dim = 1
        input_dim = 10
        hidden_dim = 64
        num_hidden_layers = 4
        agent_type = AgentType.DQN
        n_atoms = 101
        start = -100
        end = 100
        steps = 101
        model = QNetwork(input_dim=input_dim, num_hidden_layers=num_hidden_layers, hidden_layer_dim=hidden_dim,
                         agent_type=agent_type, action_space_dim=output_dim, n_atoms=n_atoms, start=start, end=end,
                         steps=steps)
        assert model is not None
        assert model.num_hidden_layers == num_hidden_layers
        assert model.input_dim == input_dim
        assert model.agent_type == agent_type
        assert model.hidden_layer_dim == hidden_dim
        assert len(list(model.network.modules())) == (num_hidden_layers + 2)
        val = model.forward(x=torch.tensor([[1.0] * input_dim]))
        assert val is not None
        assert len(val.detach().numpy()[0]) == hidden_dim



from csle_common.dao.jobs.data_collection_job_config import DataCollectionJobConfig
from csle_common.dao.jobs.system_identification_job_config import SystemIdentificationJobConfig
from csle_common.dao.jobs.training_job_config import TrainingJobConfig


class TestJobsDaoSuite:
    """
    Test suite for job data access objects (DAOs)
    """

    def test_data_collection_job_config(self, example_data_collection_job: DataCollectionJobConfig) -> None:
        """
        Tests creation and dict conversion of the DataCollectionJobConfig DAO

        :param example_data_collection_job: an example DataCollectionJobConfig
        :return: None
        """
        assert isinstance(example_data_collection_job.to_dict(), dict)
        assert isinstance(DataCollectionJobConfig.from_dict(example_data_collection_job.to_dict()),
                          DataCollectionJobConfig)
        assert (DataCollectionJobConfig.from_dict(example_data_collection_job.to_dict()).to_dict() ==
                example_data_collection_job.to_dict())
        assert (DataCollectionJobConfig.from_dict(example_data_collection_job.to_dict()) ==
                example_data_collection_job)

    def test_system_identification_job_config(
            self, example_system_identification_job_config: SystemIdentificationJobConfig) -> None:
        """
        Tests creation and dict conversion of the SystemIdentificationJobConfig DAO

        :param example_system_identification_job_config: an example SystemIdentificationJobConfig
        :return: None
        """
        assert isinstance(example_system_identification_job_config.to_dict(), dict)
        assert isinstance(SystemIdentificationJobConfig.from_dict(example_system_identification_job_config.to_dict()),
                          SystemIdentificationJobConfig)
        assert (SystemIdentificationJobConfig.from_dict(example_system_identification_job_config.to_dict()).to_dict() ==
                example_system_identification_job_config.to_dict())
        assert (SystemIdentificationJobConfig.from_dict(example_system_identification_job_config.to_dict()) ==
                example_system_identification_job_config)

    def test_training_job_config(self, example_training_job_config: TrainingJobConfig) -> None:
        """
        Tests creation and dict conversion of the TrainingJobConfig DAO

        :param example_training_job_config: an example TrainingJobConfig
        :return: None
        """
        assert isinstance(example_training_job_config.to_dict(), dict)
        assert isinstance(TrainingJobConfig.from_dict(example_training_job_config.to_dict()),
                          TrainingJobConfig)
        assert (TrainingJobConfig.from_dict(example_training_job_config.to_dict()).to_dict() ==
                example_training_job_config.to_dict())
        assert (TrainingJobConfig.from_dict(example_training_job_config.to_dict()) ==
                example_training_job_config)



from typing import List, Dict, Any
from csle_common.dao.emulation_observation.common.emulation_connection_observation_state \
    import EmulationConnectionObservationState
from csle_common.dao.emulation_config.credential import Credential
from csle_common.tunneling.forward_tunnel_thread import ForwardTunnelThread
from csle_base.json_serializable import JSONSerializable


class ConnectionSetupDTO(JSONSerializable):
    """
    DTO class containing information for setting up connections in the emulation
    """

    def __init__(self, connected: bool, credentials: List[Credential], target_connections: List[Any],
                 tunnel_threads: List[ForwardTunnelThread], forward_ports: List[int],
                 ports: List[int], interactive_shells: List[Any], non_failed_credentials: List[Credential],
                 proxies: List[EmulationConnectionObservationState], ip: str, total_time: float = 0.0):
        """
        Initializes the DTO

        :param connected: whether the connection is connected or not
        :param credentials: the list of credentials
        :param target_connections: the list of target connections
        :param tunnel_threads: the list of tunnel threads
        :param forward_ports: the list of forward ports
        :param ports: the list of ports
        :param interactive_shells: the list of interactive shells
        :param total_time: the total time of connection
        :param non_failed_credentials: the non-failed-credentials
        :param proxies: the list of proxy connections
        :param ip: the ip address of the connectio
        """
        self.connected = connected
        self.total_time = total_time
        self.credentials = credentials
        self.target_connections = target_connections
        self.tunnel_threads = tunnel_threads
        self.forward_ports = forward_ports
        self.ports = ports
        self.interactive_shells = interactive_shells
        self.total_time = total_time
        self.non_failed_credentials = non_failed_credentials
        self.proxies = proxies
        self.ip = ip

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"connected:{self.connected}, total_time:{self.total_time}, " \
               f"credentials:{list(map(lambda x: str(x), self.credentials))}, " \
               f"target_connections:{list(map(lambda x: str(x), self.target_connections))}, " \
               f"tunnel_threads:{list(map(lambda x: str(x), self.tunnel_threads))}, " \
               f"forward_ports:{list(map(lambda x: str(x), self.forward_ports))}, " \
               f"ports:{list(map(lambda x: str(x), self.ports))}, " \
               f"interactive_shells:{list(map(lambda x: str(x), self.interactive_shells))}, " \
               f"non_failed_credentials:{list(map(lambda x: str(x), self.non_failed_credentials))}, " \
               f"proxies:{list(map(lambda x: str(x), self.proxies))}, ip:{self.ip}"

    def copy(self) -> "ConnectionSetupDTO":
        """
        :return: a copy of the object
        """
        return ConnectionSetupDTO(
            connected=self.connected, credentials=self.credentials, target_connections=self.target_connections,
            tunnel_threads=self.tunnel_threads, forward_ports=self.forward_ports, ports=self.ports,
            interactive_shells=self.interactive_shells, total_time=self.total_time,
            non_failed_credentials=self.non_failed_credentials, proxies=self.proxies, ip=self.ip
        )

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "ConnectionSetupDTO":
        """
        Converts a dict representation of the object into an object instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = ConnectionSetupDTO(
            connected=d["connected"],
            credentials=list(map(lambda x: Credential.from_dict(x), d["credentials"])),
            ports=d["ports"], total_time=d["total_time"],
            non_failed_credentials=list(map(lambda x: Credential.from_dict(x), d["non_failed_credentials"])),
            proxies=list(map(lambda x: EmulationConnectionObservationState.from_dict(x), d["proxies"])),
            ip=d["ip"], forward_ports=[], interactive_shells=[], target_connections=[], tunnel_threads=[])
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation
        
        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["connected"] = self.connected
        d["credentials"] = list(map(lambda x: x.to_dict(), self.credentials))
        d["ports"] = self.ports
        d["total_time"] = self.total_time
        d["non_failed_credentials"] = list(map(lambda x: x.to_dict(), self.non_failed_credentials))
        d["ip"] = self.ip
        d["proxies"] = list(map(lambda x: x.to_dict(), self.proxies))
        return d

    @staticmethod
    def from_json_file(json_file_path: str) -> "ConnectionSetupDTO":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return ConnectionSetupDTO.from_dict(json.loads(json_str))



from unittest.mock import patch, MagicMock
from csle_common.util.import_util import ImportUtil


class TestImportUtilSuite:
    """
    Test suite for import_util
    """
    
    @patch("os.path.exists")
    @patch("csle_common.dao.system_identification.emulation_statistics.EmulationStatistics.from_json_file")
    @patch("csle_common.metastore.metastore_facade.MetastoreFacade.save_emulation_statistic")
    def test_import_emulation_statistics_from_disk_json(self, mock_save_emulation_statistic, mock_from_json_file,
                                                        mock_path_exists) -> None:
        """
        Test the method that imports emulation statistics from disk to the metastore

        :param mock_save_emulation_statistic: mock_save_emulation_statistic
        :param mock_from_json_file: mock_from_json_file
        :param mock_path_exists: mock_path_exists
        :return: None
        """
        mock_path_exists.return_value = True
        mock_statistics = MagicMock()
        mock_from_json_file.return_value = mock_statistics
        input_file = "file.json"
        emulation_name = "test_emulation"
        ImportUtil.import_emulation_statistics_from_disk_json(input_file=input_file, emulation_name=emulation_name)
        mock_path_exists.assert_called()
        mock_from_json_file.assert_called_once_with(input_file)
        assert mock_statistics.emulation_name == emulation_name
        mock_save_emulation_statistic.assert_called()

    @patch("os.path.exists")
    @patch("csle_common.dao.emulation_config.emulation_trace.EmulationTrace.load_traces_from_disk")
    @patch("csle_common.metastore.metastore_facade.MetastoreFacade.save_emulation_trace")
    def test_import_emulation_traces_from_disk_json(self, mock_save_emulation_trace, mock_load_traces_from_disk,
                                                    mock_path_exists) -> None:
        """
        Test the method that imports emulation traces from disk to the metastore

        :param mock_save_emulation_trace: mock_save_emulation_trac
        :param mock_load_traces_from_disk: mock_load_traces_from_disk
        :param mock_path_exists: mock_path_exists
        :return: None
        """
        mock_path_exists.return_value = True
        mock_trace_1 = MagicMock()
        mock_trace_2 = MagicMock()
        mock_load_traces_from_disk.return_value = [mock_trace_1, mock_trace_2]
        input_file = "file.json"
        emulation_name = "test_emulation"
        ImportUtil.import_emulation_traces_from_disk_json(input_file=input_file, emulation_name=emulation_name)
        mock_path_exists.assert_called()
        mock_load_traces_from_disk.assert_called()
        assert mock_trace_1.emulation_name == emulation_name
        assert mock_trace_2.emulation_name == emulation_name
        assert mock_save_emulation_trace.call_count == 2



from typing import List, Dict, Union, Any
import numpy as np
from numpy.typing import NDArray
import iteround
from csle_common.dao.training.policy import Policy
from csle_common.dao.training.agent_type import AgentType
from csle_common.dao.training.player_type import PlayerType
from csle_common.dao.simulation_config.state import State
from csle_common.dao.simulation_config.action import Action
from csle_common.dao.training.experiment_config import ExperimentConfig
from csle_common.dao.training.policy_type import PolicyType
from csle_common.dao.training.linear_tabular_policy import LinearTabularPolicy


class MixedLinearTabularPolicy(Policy):
    """
    A mixed policy using an ensemble of linear tabulat policies
    """

    def __init__(self, simulation_name: str, player_type: PlayerType, states: List[State],
                 actions: List[Action], experiment_config: ExperimentConfig, avg_R: float):
        """
        Initializes the policy

        :param simulation_name: the simulation name
        :param states: list of states (required for computing stage policies)
        :param actions: list of actions
        :param experiment_config: the experiment configuration for training the policy
        :param avg_R: the average reward of the policy when evaluated in the simulation
        """
        super(MixedLinearTabularPolicy, self).__init__(agent_type=AgentType.DFSP_LOCAL, player_type=player_type)
        self.linear_tabular_policies: List[LinearTabularPolicy] = []
        self.id = -1
        self.simulation_name = simulation_name
        self.states = states
        self.actions = actions
        self.experiment_config = experiment_config
        self.avg_R = avg_R
        self.policy_type = PolicyType.MIXED_LINEAR_TABULAR

    def action(self, o: List[float], deterministic: bool = True) -> Union[int, List[int], float, NDArray[Any]]:
        """
        Multi-threshold stopping policy

        :param o: the current observation
        :param deterministic: boolean flag indicating whether the action selection should be deterministic
        :return: the selected action
        """
        policy = np.random.choice(self.linear_tabular_policies)
        a = policy.action(o=o)
        return float(a)

    def probability(self, o: List[float], a: int) -> int:
        """
        Probability of a given action

        :param o: the current observation
        :param a: a given action
        :return: the probability of a
        """
        return self.action(o=o) == a

    def to_dict(self) -> Dict[str, Any]:
        """
        :return: a dict representation of the policy
        """
        d: Dict[str, Any] = {}
        d["id"] = self.id
        d["simulation_name"] = self.simulation_name
        d["linear_tabular_policies"] = list(map(lambda x: x.to_dict(), self.linear_tabular_policies))
        d["states"] = list(map(lambda x: x.to_dict(), self.states))
        d["player_type"] = self.player_type
        d["actions"] = list(map(lambda x: x.to_dict(), self.actions))
        d["experiment_config"] = self.experiment_config.to_dict()
        d["agent_type"] = self.agent_type
        d["avg_R"] = self.avg_R
        d["policy_type"] = self.policy_type
        return d

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "MixedLinearTabularPolicy":
        """
        Converst a dict representation of the object to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        linear_tabular_policies = list(map(lambda x: x.from_dict(), d["linear_tabular_policies"]))
        obj = MixedLinearTabularPolicy(simulation_name=d["simulation_name"],
                                       states=list(map(lambda x: State.from_dict(x), d["states"])),
                                       player_type=d["player_type"],
                                       actions=list(map(lambda x: Action.from_dict(x), d["actions"])),
                                       experiment_config=ExperimentConfig.from_dict(d["experiment_config"]),
                                       avg_R=d["avg_R"])
        obj.linear_tabular_policies = linear_tabular_policies
        obj.id = d["id"]
        return obj

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"linear_tabular_policies: {self.linear_tabular_policies}, id: {self.id}, " \
               f"simulation_name: {self.simulation_name}, " \
               f"states: {self.states}, experiment_config: {self.experiment_config}," \
               f"avg_R: {self.avg_R}, policy type: {self.policy_type}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "MixedLinearTabularPolicy":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return MixedLinearTabularPolicy.from_dict(json.loads(json_str))

    def copy(self) -> "MixedLinearTabularPolicy":
        """
        :return: a copy of the DTO
        """
        return self.from_dict(self.to_dict())

    def stage_policy(self, o: Union[List[Union[int, float]], int, float]) -> List[List[float]]:
        """
        Returns the stage policy for a given observation

        :param o: the observation to return the stage policy for
        :return: the stage policy
        """
        stage_policies = []
        for policy in self.linear_tabular_policies:
            stage_policies.append(policy.stage_policy(o=o))
        stage_strategy = np.zeros((len(self.states), len(self.actions)))
        for i, s_a in enumerate(self.states):
            for j, a in enumerate(self.actions):
                stage_strategy[i][j] = sum([stage_policies[k][i][j]
                                            for k in range(len(stage_policies))]) / len(stage_policies)
            stage_strategy[i] = iteround.saferound(stage_strategy[i], 2)
        return list(stage_strategy.tolist())
